@@ -1,5 +1,15 @@
+from enum import IntEnum
+from types import MappingProxyType
+
+
+class CFFOptimization(IntEnum):
+    NONE = 0
+    SPECIALIZE = 1
+    SUBROUTINIZE = 2
+
+
 SPARSE_TTF_MASTER_TABLES = frozenset(
-    ["glyf", "head", "hmtx", "loca", "maxp", "post", "vmtx"]
+    ["glyf", "head", "hmtx", "loca", "maxp", "post", "vmtx", "cvt ", "fpgm", "prep"]
 )
 SPARSE_OTF_MASTER_TABLES = frozenset(["CFF ", "VORG", "head", "hmtx", "maxp", "vmtx"])
 
@@ -33,14 +43,30 @@ COLOR_LAYER_MAPPING_KEY = UFO2FT_PREFIX + "colorLayerMapping"
 # ]
 COLR_CLIP_BOXES_KEY = UFO2FT_PREFIX + "colrClipBoxes"
 
+GLYPHS_MATH_PREFIX = "com.nagwa.MATHPlugin."
+GLYPHS_MATH_CONSTANTS_KEY = GLYPHS_MATH_PREFIX + "constants"
+GLYPHS_MATH_VARIANTS_KEY = GLYPHS_MATH_PREFIX + "variants"
+GLYPHS_MATH_EXTENDED_SHAPE_KEY = GLYPHS_MATH_PREFIX + "extendedShape"
+
+OBJECT_LIBS_KEY = "public.objectLibs"
 OPENTYPE_CATEGORIES_KEY = "public.openTypeCategories"
 OPENTYPE_META_KEY = "public.openTypeMeta"
-
+OPENTYPE_POST_UNDERLINE_POSITION_KEY = "public.openTypePostUnderlinePosition"
+TRUETYPE_INSTRUCTIONS_KEY = "public.truetype.instructions"
+TRUETYPE_METRICS_KEY = "public.truetype.useMyMetrics"
+TRUETYPE_OVERLAP_KEY = "public.truetype.overlap"
+TRUETYPE_ROUND_KEY = "public.truetype.roundOffsetToGrid"
 UNICODE_VARIATION_SEQUENCES_KEY = "public.unicodeVariationSequences"
 
+COMMON_SCRIPT = "Zyyy"
+
+UNICODE_SCRIPT_ALIASES = MappingProxyType({"Hira": "Hrkt", "Kana": "Hrkt"})
+
+
+# HarfBuzz passes Sinhala to the Indic shaper, while OpenType moved it to the USE shaper.
+
 INDIC_SCRIPTS = [
     "Beng",  # Bengali
-    "Cham",  # Cham
     "Deva",  # Devanagari
     "Gujr",  # Gujarati
     "Guru",  # Gurmukhi
@@ -53,80 +79,81 @@ INDIC_SCRIPTS = [
 ]
 
 USE_SCRIPTS = [
-    # Correct as at Unicode 14.0
-    "Tibt",  # Tibetan
-    "Mong",  # Mongolian
-    # HB has Sinhala commented out here?!
-    "Buhd",  # Buhid
-    "Hano",  # Hanunoo
-    "Tglg",  # Tagalog
-    "Tagb",  # Tagbanwa
-    "Limb",  # Limbu
-    "Tale",  # Tai Le
-    "Bugi",  # Buginese
-    "Khar",  # Kharosthi
-    "Sylo",  # Syloti Nagri
-    "tfng",  # Tifinagh
+    # Correct as at Unicode 15.0
+    "Adlm",  # Adlam
+    "Ahom",  # Ahom
     "Bali",  # Balinese
-    "Nko ",  # Nko
-    "Phag",  # Phags Pa
-    "Cham",  # Cham
-    "Kali",  # Kayah Li
-    "Lepc",  # Lepcha
-    "Rjng",  # Rejang
-    "Saur",  # Saurashtra
-    "Sund",  # Sundanese
-    "Egyp",  # Egyptian Hieroglyphs
-    "Java",  # Javanese
-    "Kthi",  # Kaithi
-    "Mtei",  # Meetei Mayek
-    "Lana",  # Tai Tham
-    "Tavt",  # Tai Viet
     "Batk",  # Batak
     "Brah",  # Brahmi
-    "Mand",  # Mandaic
+    "Bugi",  # Buginese
+    "Buhd",  # Buhid
     "Cakm",  # Chakma
-    "Plrd",  # Miao
-    "Shrd",  # Sharada
-    "Takr",  # Takri
-    "Dupl",  # Duployan
-    "Gran",  # Grantha
-    "Khoj",  # Khojki
-    "Sind",  # Khudawadi
-    "Mahj",  # Mahajani
-    "Mani",  # Manichaean
-    "Modi",  # Modi
-    "Hmng",  # Pahawh Hmong
-    "Phlp",  # Psalter Pahlavi
-    "Sidd",  # Siddham
-    "Tirh",  # Tirhuta
-    "Ahom",  # Ahom
-    "Mult",  # Multani
-    "Adlm",  # Adlam
-    "Nhks",  # Bhaiksuki
-    "Marc",  # Marchen
-    "Newa",  # Newa
-    "Gonm",  # Masaram Gondi
-    "Soyo",  # Soyombo
-    "Zanb",  # Zanabazar Square
-    "Dogr",  # Dogra
-    "Gong",  # Gunjala Gondi
-    "Rohg",  # Hanifi Rohingya
-    "Maka",  # Makasar
-    "Medf",  # Medefaidrin
-    "Sogo",  # Old Sogdian
-    "Sogd",  # Sogdian
-    "Elym",  # Elymaic
-    "Nand",  # Nandinagari
-    "Hmnp",  # Nyiakeng Puachue Hmong
-    "Wcho",  # Wancho
+    "Cham",  # Cham
     "Chrs",  # Chorasmian
-    "Diak",  # Dives Akuru
-    "Kits",  # Khitan Small Script
-    "Yezi",  # Yezidi
     "Cpmn",  # Cypro Minoan
+    "Diak",  # Dives Akuru
+    "Dogr",  # Dogra
+    "Dupl",  # Duployan
+    "Egyp",  # Egyptian Hieroglyphs
+    "Elym",  # Elymaic
+    "Gong",  # Gunjala Gondi
+    "Gonm",  # Masaram Gondi
+    "Gran",  # Grantha
+    "Hano",  # Hanunoo
+    "Hmng",  # Pahawh Hmong
+    "Hmnp",  # Nyiakeng Puachue Hmong
+    "Java",  # Javanese
+    "Kali",  # Kayah Li
+    "Kawi",  # Kawi
+    "Khar",  # Kharosthi
+    "Khoj",  # Khojki
+    "Kits",  # Khitan Small Script
+    "Kthi",  # Kaithi
+    "Lana",  # Tai Tham
+    "Lepc",  # Lepcha
+    "Limb",  # Limbu
+    "Mahj",  # Mahajani
+    "Maka",  # Makasar
+    "Mand",  # Mandaic
+    "Mani",  # Manichaean
+    "Marc",  # Marchen
+    "Medf",  # Medefaidrin
+    "Modi",  # Modi
+    "Mong",  # Mongolian
+    "Mtei",  # Meetei Mayek
+    "Mult",  # Multani
+    "Nagm",  # Nag Mundari
+    "Nand",  # Nandinagari
+    "Newa",  # Newa
+    "Nhks",  # Bhaiksuki
+    "Nko ",  # Nko
     "Ougr",  # Old Uyghur
+    "Phag",  # Phags Pa
+    "Phlp",  # Psalter Pahlavi
+    "Plrd",  # Miao
+    "Rjng",  # Rejang
+    "Rohg",  # Hanifi Rohingya
+    "Saur",  # Saurashtra
+    "Shrd",  # Sharada
+    "Sidd",  # Siddham
+    "Sind",  # Khudawadi
+    "Sogd",  # Sogdian
+    "Sogo",  # Old Sogdian
+    "Soyo",  # Soyombo
+    "Sund",  # Sundanese
+    "Sylo",  # Syloti Nagri
+    "Tagb",  # Tagbanwa
+    "Takr",  # Takri
+    "Tale",  # Tai Le
+    "Tavt",  # Tai Viet
+    "Tfng",  # Tifinagh
+    "Tglg",  # Tagalog
+    "Tibt",  # Tibetan
+    "Tirh",  # Tirhuta
     "Tnsa",  # Tangsa
     "Toto",  # Toto
     "Vith",  # Vithkuqi
+    "Wcho",  # Wancho
+    "Yezi",  # Yezidi
+    "Zanb",  # Zanabazar Square
 ]
