@@ -6,12 +6,14 @@ from ufo2ft.constants import FEATURE_WRITERS_KEY
 from ufo2ft.util import _loadPluginFromString
 
 from .baseFeatureWriter import BaseFeatureWriter
+from .cursFeatureWriter import CursFeatureWriter
 from .gdefFeatureWriter import GdefFeatureWriter
 from .kernFeatureWriter import KernFeatureWriter
 from .markFeatureWriter import MarkFeatureWriter
 
 __all__ = [
     "BaseFeatureWriter",
+    "CursFeatureWriter",
     "GdefFeatureWriter",
     "KernFeatureWriter",
     "MarkFeatureWriter",
