@@ -4,7 +4,6 @@ from ufo2ft.util import _GlyphSet
 
 
 class ExplodeColorLayerGlyphsFilter(BaseFilter):
-
     """This filter doesn't really filter glyphs, but copies glyphs
     from UFO layers to alternate glyphs in the default layer, for use
     in the COLR table.
@@ -48,6 +47,10 @@ class ExplodeColorLayerGlyphsFilter(BaseFilter):
                 layerGlyphSet, glyphSet, component.baseGlyph, layerName
             )
             component.baseGlyph = baseLayerGlyphName
+        # The new alternates must have their codepoints stripped lest they
+        # become encoded, which they will be if placed in the default layer.
+        # See: https://github.com/googlefonts/ufo2ft/pull/739#issuecomment-1516075892
+        layerGlyph.unicodes = []
         glyphSet[layerGlyphName] = layerGlyph
         self.context.colorLayerGlyphNames.add(layerGlyphName)
         return layerGlyphName
