@@ -15,8 +15,6 @@ from ufo2ft.featureWriters import (
     ast,
 )
 
-from .testSupport import pushd
-
 
 class ParseLayoutFeaturesTest:
     def test_include(self, FontClass, tmpdir):
@@ -40,14 +38,16 @@ class ParseLayoutFeaturesTest:
 
         assert "# hello world" in str(fea)
 
-    def test_include_no_ufo_path(self, FontClass, tmpdir):
+    def test_include_no_ufo_path(self, FontClass, tmpdir, monkeypatch):
         ufo = FontClass()
         ufo.features.text = dedent(
             """\
             include(test.fea)
             """
         )
-        with pushd(str(tmpdir)):
+        with monkeypatch.context() as context:
+            context.chdir(str(tmpdir))
+            ufo.save("Test.ufo")
             with pytest.raises(IncludedFeaNotFound):
                 parseLayoutFeatures(ufo)
 
@@ -100,6 +100,18 @@ class ParseLayoutFeaturesTest:
 
         assert "# hello world" in str(fea)
 
+    def test_include_dir_cwd(self, FontClass, tmp_path, monkeypatch):
+        (tmp_path / "test.fea").write_text("# hello world", encoding="utf-8")
+        ufo = FontClass()
+        ufo.features.text = "include(test.fea)"
+
+        with monkeypatch.context() as context:
+            context.chdir(tmp_path)
+            ufo.save("Test.ufo")
+            fea = parseLayoutFeatures(ufo)
+
+        assert "# hello world" in str(fea)
+
 
 class DummyFeatureWriter:
     tableTag = "GPOS"
@@ -224,14 +236,17 @@ class FeatureCompilerTest:
 
     def test_GSUB_writers_run_first(self, FontClass):
         class FooFeatureWriter(BaseFeatureWriter):
-
             tableTag = "GSUB"
 
             def write(self, font, feaFile, compiler=None):
                 foo = ast.FeatureBlock("FOO ")
                 foo.statements.append(
                     ast.SingleSubstStatement(
-                        "a", "v", prefix="", suffix="", forceChain=None
+                        [ast.GlyphName("a")],
+                        [ast.GlyphName("v")],
+                        prefix="",
+                        suffix="",
+                        forceChain=None,
                     )
                 )
                 feaFile.statements.append(foo)
@@ -326,14 +341,15 @@ class FeatureCompilerTest:
                 sub f f by f_f;
             } liga;
 
-
-            lookup kern_ltr {
+            lookup kern_Default {
                 lookupflag IgnoreMarks;
                 pos a v -40;
-            } kern_ltr;
+            } kern_Default;
 
             feature kern {
-                lookup kern_ltr;
+                script DFLT;
+                language dflt;
+                lookup kern_Default;
             } kern;
             """
         )
